@@ -1,5 +1,3 @@
-#!/usr/bin/python
-
 ''' Python API for YouTube
     Copyright (C)  2013 nagev
 
@@ -16,7 +14,7 @@
     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.  '''
 
-__version__ = "0.3.07"
+__version__ = "0.3.08"
 __author__ = "nagev"
 __license__ = "GPLv3"
 
@@ -38,47 +36,45 @@ else:
     from urllib import unquote_plus
     from urlparse import parse_qs
 
-#logging.basicConfig(level=logging.DEBUG)
+logging.basicConfig(level=logging.DEBUG)
 
-def _extract_function_from_js(funcname, js):
-    # Find a function definition named funcname and extract components
-    match = re.search(r'function %s\(((?:\w+,?)+)\)\{([^\{]+)\}' % funcname, js)
-    return {'name': funcname, 'argnames': match.group(1).split(","),
-        'body': match.group(2) }
+def _extract_function_from_js(name, js):
+    # Find a function called `name` and extract components
+    m = re.search(r'function %s\(((?:\w+,?)+)\)\{([^}]+)\}' % name, js)
+    return {'name': name, 'parameters': m.group(1).split(","),
+        'body': m.group(2) }
 
 def _getval(val, argsdict): # resolves variable values. preserves int literals
-    match = re.match(r'(\d+)', val)
-    if match:
-        return(int(match.group(1)))
+    m = re.match(r'(\d+)', val)
+    if m:
+        return(int(m.group(1)))
     elif val in argsdict:
         return argsdict[val]
     else:
         raise RuntimeError("Error val %s from dict %s" % (val, argsdict))
 
-def _get_func_from_call(f_old, fname, argscall, js):
-    argscall = argscall.split(",")
-    newfunc = _extract_function_from_js(fname, js)
-    newfunc['args'] = {}
-    for n, argname in enumerate(argscall):
-        value = _getval(argname, f_old['args'])
-        # curvar is the argument name specified in the new function definition
-        curvar = newfunc['argnames'][n]
-        newfunc['args'][curvar] = value
-    return newfunc
+def _get_func_from_call(caller_function, name, arguments, js):
+    newfunction = _extract_function_from_js(name, js)
+    newfunction['args'] = {}
+    for n, arg in enumerate(arguments):
+        value = _getval(arg, caller_function['args'])
+        param = newfunction['parameters'][n]
+        newfunction['args'][param] = value
+    return newfunction
 
 def _solve(f, js):
     # solve basic javascript function
     parts = f['body'].split(";")
     for part in parts:
         logging.debug("Working on part: %s" % part)
-        # split, do nothing
-        m = re.match(r'(\w+)=(\w+)\.split\(""\)', part)
+        # split or join, do nothing
+        m = re.match(r'(\w+)=(\w+)\.(?:split|join)\(""\)', part)
         if m and m.group(1) == m.group(2):
             continue
         m = re.match(r'(\w+)=(\w+)\(((?:\w+,?)+)\)', part)
         if m: # a function call
-            lhs, fname, argscall = m.group(*range(1,4))
-            newfunc = _get_func_from_call(f, fname, argscall, js)
+            lhs, funcname, args = m.group(*range(1,4))
+            newfunc = _get_func_from_call(f, funcname, args.split(","), js)
             f['args'][lhs] = _solve(newfunc, js) # recursive call
             continue
         m = re.match(r'var\s(\w+)=(\w+)\[(\w+)\]', part)
@@ -113,15 +109,13 @@ def _solve(f, js):
 
 def _decodesig(sig, js):
     # get main function name from a function call
-    sigargument = "g.s"
-    sigprefix = "g.sig"
-    match = re.search(r'%s\|\|(\w+)\(%s\)' % (sigprefix, sigargument), js)
-    funcname = match.group(1)
+    m = re.search(r'\w\.sig\|\|(\w+)\(\w+\.\w+\)', js)
+    funcname = m.group(1)
     function = _extract_function_from_js(funcname, js)
-    if not len(function['argnames']) == 1:
+    if not len(function['parameters']) == 1:
         raise RuntimeError("Main sig js function has more than one arg: %s" %
-            function['argnames'])
-    function['args'] = {function['argnames'][0]: sig}
+            function['parameters'])
+    function['args'] = {function['parameters'][0]: sig}
     return _solve(function, js)
 
 class Stream():
@@ -146,7 +140,7 @@ class Stream():
     
     def __init__(self, streammap, opener, title="ytvid", js=None):
         if not streammap.get("sig", ""):
-            logging.debug("Decrypting sig: %s" % streammap['s'])
+            logging.debug("Decrypting sig: %s" % streammap['s'][0])
             streammap['sig'] = [_decodesig(streammap['s'][0], js)]
             logging.debug("Calculated decrypted sig: %s" % streammap['sig'][0])
         self.url = streammap['url'][0] + '&signature=' + streammap['sig'][0]
@@ -156,37 +150,39 @@ class Stream():
         self.itag = streammap['itag'][0]
         self.title = title
         self.filename = self.title + "." + self.extension
+        self.filesize = None
         self._opener = opener
 
     def get_filesize(self):
-        opener = self._opener
-        return int(opener.open(self.url).headers['content-length'])
+        if not self.filesize:
+            opener = self._opener
+            self.filesize = int(opener.open(self.url).headers['content-length'])
+        return self.filesize
 
-    def download(self, progress=True, filepath=""):
-        response = self._opener.open(self.url)
-        total = int(response.info()['Content-Length'].strip())
-        print("-Downloading '{}' [{:,} Bytes]".format(self.filename,
-            total))
+    def download(self, filepath="", quiet=False, callback=None):
         status_string = ('  {:,} Bytes [{:.2%}] received. Rate: [{:4.0f} '
                          'kbps].  ETA: [{:.0f} secs]')
-        chunksize, bytesdone, t0 = 1024, 0, time.time()
+        response = self._opener.open(self.url)
+        total = int(response.info()['Content-Length'].strip())
+        chunksize, bytesdone, t0 = 16384, 0, time.time()
         outfh = open(filepath or self.filename, 'wb')
-        while 1:
+        while True:
             chunk = response.read(chunksize)
-            elapsed = time.time() - t0
             outfh.write(chunk)
+            elapsed = time.time() - t0
             bytesdone += len(chunk)
+            rate = (bytesdone / 1024) / elapsed
+            eta = (total - bytesdone) / (rate * 1024)
+            progress_stats = (bytesdone, bytesdone * 1.0 / total, rate, eta)
             if not chunk:
                 outfh.close()
                 break
-            if progress:
-                rate = (bytesdone / 1024) / elapsed
-                eta = (total - bytesdone) / (rate * 1024)
-                display = (bytesdone, bytesdone * 1.0 / total, rate, eta)
-                status = status_string.format(*display)
+            if not quiet:
+                status = status_string.format(*progress_stats)
                 sys.stdout.write("\r" + status + ' ' * 4 + "\r")
                 sys.stdout.flush()
-        print("\nDone")
+            if callback:
+                callback(total, *progress_stats)
 
 
 class Pafy():
@@ -195,7 +191,6 @@ class Pafy():
         return self.length
 
     def __repr__(self):
-        out = ""
         keys = "Title Author ID Duration Rating Views Thumbnail Keywords"
         keys = keys.split(" ")
         keywords = ", ".join(self.keywords)
@@ -208,38 +203,37 @@ class Pafy():
                     ID=self.videoid,
                     Thumbnail=self.thumb,
                     Keywords=keywords)
-        for k in keys: 
-            try:
-                out += "%s: %s\n" % (k, info[k])
-            except KeyError:
-                pass
-        return out
+        return "\n".join(["%s: %s" % (k, info.get(k, "")) for k in keys])
 
     def _setmetadata(self, allinfo):
-        self.title = allinfo['title'][0]
-        self.author = allinfo['author'][0]
-        self.videoid = allinfo['video_id'][0]
-        self.rating = float(allinfo['avg_rating'][0])
-        self.length = int(allinfo['length_seconds'][0])
+        f = lambda x: allinfo[x][0]
+        self.title = f('title')
+        self.author = f('author')
+        self.videoid = f('video_id')
+        self.rating = float(f('avg_rating'))
+        self.length = int(f('length_seconds'))
+        self.viewcount = int(f('view_count'))
+        self.thumb = unquote_plus(f('thumbnail_url'))
         self.duration = time.strftime('%H:%M:%S', time.gmtime(self.length))
-        self.viewcount = int(allinfo['view_count'][0])
-        self.thumb = unquote_plus(allinfo['thumbnail_url'][0])
-        self.formats = allinfo['fmt_list'][0].split(",")
+        self.formats = f('fmt_list').split(",")
         self.formats = [x.split("/") for x in self.formats]
         if 'keywords' in allinfo:
-            self.keywords = allinfo['keywords'][0].split(',')
+            self.keywords = f('keywords').split(',')
         if allinfo.get('iurlsd'):
-            self.bigthumb = allinfo['iurlsd'][0]
+            self.bigthumb = f('iurlsd')
         if allinfo.get('iurlmaxres'):
-            self.bigthumbhd = allinfo['iurlmaxres'][0]
-        return 
+            self.bigthumbhd = f('iurlmaxres')
+        return
 
     def __init__(self, video_url):
         infoUrl = 'https://www.youtube.com/get_video_info?video_id='
-        try:
-            vidid = re.search(r'v=([a-zA-Z0-9-_]*)', video_url).group(1)
-        except:
-            raise RuntimeError("bad video url")
+        m = re.search(r'\bv=([a-zA-Z0-9-_]{11})', video_url)
+        if not m:
+            m = re.search(r'\b([a-zA-Z0-9-_]{11})', video_url)
+        if not m:
+            error = "Need 11 character video id or the URL of the video. Got %s"
+            raise RuntimeError(error % video_url)
+        vidid = m.group(1)
         infoUrl += vidid + "&asv=3&el=detailpage&hl=en_US"
         opener = build_opener()
         ua = ("Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64;"
@@ -247,6 +241,9 @@ class Pafy():
         opener.addheaders = [('User-Agent', ua)]
         self.keywords = ""
         allinfo = parse_qs(decode_if_py3(opener.open(infoUrl).read()))
+        if allinfo['status'][0] == "fail":
+            reason = allinfo['reason'][0] or "Bad video argument"
+            raise RuntimeError("Youtube says: %s" % reason)
         self._setmetadata(allinfo)
         streamMap = allinfo['url_encoded_fmt_stream_map'][0].split(',')
         smap = [parse_qs(sm) for sm in streamMap]
@@ -254,9 +251,9 @@ class Pafy():
         if not smap[0].get("sig", ""):  # vevo!
             watchurl = "https://www.youtube.com/watch?v=" + vidid
             watchinfo = opener.open(watchurl).read().decode("UTF-8")
-            match = re.search(r';ytplayer.config = ({.*?});', watchinfo)
+            m = re.search(r';ytplayer.config = ({.*?});', watchinfo)
             try:
-                myjson = json.loads(match.group(1))
+                myjson = json.loads(m.group(1))
             except:
                 raise RuntimeError('Problem handling this video')
             args = myjson['args']
